import numpy as np
import pandas as pd
import time


from envs import (
    Environment,
    NoiseGenerator,
    StochasticContextGenerator as CtxGenerator,
    Example2ContextGenerator as Ex2CtxGenerator,
    Example1ContextGenerator as Ex1CtxGenerator,
    RealDatasetEnvironment,
    RealDatasetContextGenerator,
    RealDatasetLabelGenerator,
)
from envs import GroupedStochasticContextGenerator as GroupedCtxGenerator

from policies import Roful

from scipy.io import arff
from sklearn.preprocessing import StandardScaler

from plotting import *
from utils import StateFactory, save_results


def load_dataset(dataset, scaler):
    """Load and preprocess a dataset"""
    data, meta = arff.loadarff(f"datasets/{dataset}.arff")
    df = pd.DataFrame(data).dropna().sample(frac=1).reset_index(drop=True)

    if dataset == "JapaneseVowels":
        features = scaler.fit_transform(df.iloc[:, 1:])
        labels = df.iloc[:, 0]
    else:
        features = scaler.fit_transform(df.iloc[:, :-1])
        labels = df.iloc[:, -1]

    return features, labels


def get_dataset_k(dataset):
    """Get number of classes for each dataset"""
    k_map = {"cardiotocography": 10, "segment": 7, "JapaneseVowels": 9}
    return k_map[dataset]


def setup_algorithms(
    d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS=10.0, mu_Greedy=12.0
):
    """Initialize all algorithms"""
    return {
        "TS-MR": Roful.tsmr(
            d,
            lambda_=lambda_,
            state=state_factory(),
            inflation=1.0,
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
            mu_threshold=mu_TS,
            radius_oful=Roful.radius_OFUL(),
        ),
        "Greedy-MR": Roful.tsmr(
            d,
            lambda_=lambda_,
            state=state_factory(),
            inflation=0.0,
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
            mu_threshold=mu_Greedy,
            radius_oful=Roful.radius_OFUL(),
        ),
        "LinTS": Roful.ts(
            d,
            lambda_=lambda_,
            state=state_factory(),
            inflation=1.0,
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
        ),
        "TS-Freq": Roful.ts(
            d,
            lambda_=lambda_,
            state=state_factory(),
            inflation=Roful.ts_freq_inflation(),
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
        ),
        "OFUL": Roful.oful(
            d,
            lambda_=lambda_,
            radius_oful=Roful.radius_OFUL(),
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
        ),
        "Greedy": Roful.ts(
            d,
            lambda_=lambda_,
            state=state_factory(),
            inflation=0.0,
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=1e-3,
        ),
    }


def execute_experiment(env, algs, t):
    """Run experiment for t iterations"""
    for i in range(t):
        ctx = env.next()
        for alg in algs.values():
            idx = alg.choose_arm(ctx)
            fb = env.get_feedback(idx)
            alg.update(fb)

        if i % 100 == 0:
            print(f"iteration {i} of {t} done")

    return {name: alg.outputs for name, alg in algs.items()}


def emp_scenario(
    state_factory,
    k=2,
    d=300,
    t=1000,
    prior_var=1.0,
    prior_mu=0.0,
    noise_sd=1.0,
    radius=1.0,
    delta=0.1,
    mu_TS=10.0,
    mu_Greedy=12.0,
    dataset="eeg",
):

    scaler = StandardScaler()
    features, labels = load_dataset(dataset, scaler)
    k = get_dataset_k(dataset)
    d = k * features.shape[1]

    print(f"d = {d}, k = {k}, t = {t}")

    ctx_gen = RealDatasetContextGenerator(k, d, features)
    noise_gen = NoiseGenerator.gaussian_noise(noise_sd, state=state_factory())
    label_gen = RealDatasetLabelGenerator(d, labels)
    env = RealDatasetEnvironment(ctx_gen, label_gen, labels.unique(), noise_gen)

    lambda_ = prior_var
    param = 0.0001 * np.ones(d)
    algs = setup_algorithms(
        d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS, mu_Greedy
    )

    return execute_experiment(env, algs, t)


def example1_scenario(
    state_factory,
    k=3,
    d=300,
    t=1000,
    prior_var=1.0,
    prior_mu=0.0,
    noise_sd=1.0,
    radius=1.0,
    delta=0.1,
    mu_TS=10.0,
    mu_Greedy=12.0,
):

    param = state_factory().normal(prior_mu, prior_var**0.5, d)
    radius = (d * prior_var) ** 0.5
    lambda_ = 1

    ctx_gen = Ex1CtxGenerator(d, np.sign(noise_sd**2 - prior_var))
    noise_gen = NoiseGenerator.gaussian_noise(noise_sd, state=state_factory())
    env = Environment(param, ctx_gen, noise_gen)

    algs = setup_algorithms(
        d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS, mu_Greedy
    )

    return execute_experiment(env, algs, t)


def ex3_scenario(
    state_factory,
    k=3,
    d=300,
    t=1000,
    prior_var=1.0,
    prior_mu=0.0,
    noise_sd=1.0,
    radius=1.0,
    delta=0.1,
    mu_TS=12.0,
    mu_Greedy=12.0,
):

    if k != 3:
        print("*** Warning ***, k should be 3 for fixed action case")

    param = state_factory().normal(prior_mu, prior_var**0.5, d)
    radius = (d * prior_var) ** 0.5
    lambda_ = noise_sd**2 / prior_var

    ctx_gen = Ex2CtxGenerator(d)
    noise_gen = NoiseGenerator.gaussian_noise(noise_sd, state=state_factory())
    env = Environment(param, ctx_gen, noise_gen)

    algs = setup_algorithms(
        d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS, mu_Greedy
    )

    return execute_experiment(env, algs, t)


def ex2_scenario(
    state_factory,
    k=2,
    d=300,
    t=1000,
    prior_var=1.0,
    prior_mu=0.0,
    noise_sd=1.0,
    radius=70.0,
    delta=0.1,
    mu_TS=12.0,
    mu_Greedy=12.0,
    g=None,
):

    param = state_factory().randn(d) * prior_var**0.5
    param = radius * param / np.linalg.norm(param, ord=2)
    lambda_ = noise_sd**2 / prior_var

    ctx_gen = GroupedCtxGenerator(k, d, grouped=g, state=state_factory())
    noise_gen = NoiseGenerator.gaussian_noise(noise_sd, state=state_factory())
    env = Environment(param, ctx_gen, noise_gen)

    algs = setup_algorithms(
        d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS, mu_Greedy
    )

    return execute_experiment(env, algs, t)


def ex1_scenario(
    state_factory,
    k=1,
    d=100,
    t=1000,
    prior_var=2.0,
    prior_mu=0.0,
    noise_sd=1.0,
    radius=1.0,
    delta=0.1,
    mu_TS=8.0,
    mu_Greedy=8.0,
):

    param = state_factory().normal(prior_mu, prior_var**0.5, d)
    param = radius * param / np.linalg.norm(param, ord=2)
    prior_var = radius**2 / d
    lambda_ = noise_sd**2 / prior_var

    k = int(k)
    ctx_gen = CtxGenerator.uniform_on_sphere(k, d, np.sqrt(d), state=state_factory())
    noise_gen = NoiseGenerator.gaussian_noise(noise_sd, state=state_factory())
    env = Environment(param, ctx_gen, noise_gen)

    algs = setup_algorithms(
        d, lambda_, state_factory, param, noise_sd, t, radius, mu_TS, mu_Greedy
    )

    return execute_experiment(env, algs, t)


def run_simulation(
    n,
    k,
    d,
    t,
    s,
    prior_mu=0.0,
    prior_sd=10.0,
    noise_sd=1.0,
    sim=None,
    radius=1.0,
    mu_TS=None,
    mu_Greedy=None,
    dataset=None,
):
    state_factory = StateFactory(s)
    saver = save_results()

    timestr = time.strftime("%Y%m%d-%H%M%S")
    output_name = (
        f"sim_{sim}_n_{n}_k_{k}_t_{t}_r_{radius}"
        f"_pm_{prior_mu}_ps_{prior_sd}_ns_{noise_sd}"
        f"_mts_{mu_TS}_mg_{mu_Greedy}_ds_{dataset}_{timestr}"
    )
    output_folder = "my_outputs"
    saver.init_outputs()

    for i in range(n):
        print(f"Running experiment {i} ...")

        scenario_funcs = {
            "geo_aware_ex1": lambda **kwargs: ex1_scenario(**kwargs, d=d, k=k),
            "geo_aware_ex2": lambda **kwargs: ex2_scenario(**kwargs, d=50, k=10, g=1),
            "geo_aware_ex3": lambda **kwargs: ex3_scenario(**kwargs, d=d, k=k),
            "geo_aware_emp": lambda **kwargs: emp_scenario(**kwargs, dataset=dataset),
        }

        scenario = scenario_funcs.get(sim)
        if not scenario:
            raise ValueError(f"Invalid simulation type: {sim}")

        results = scenario(
            t=t,
            state_factory=state_factory,
            prior_var=prior_sd**2,
            prior_mu=prior_mu,
            noise_sd=noise_sd,
            radius=radius,
            mu_TS=mu_TS,
            mu_Greedy=mu_Greedy,
        )

        saver.aggregate_metrics(results)

    metrics, output, output_last = saver.aggregate_output()
    outputs, outputs_last = saver.save_outputs(output_folder, output_name)

    figure_folder = f"figures/figures-{output_name}"
    figure_suffix = f".jpg"
    plot_output(metrics, output, saver.labels, figure_folder, figure_suffix, dataset)

    outputs.to_csv(f"{figure_folder}/all-{output_name}.csv", index=False)
    outputs_last.to_csv(f"{figure_folder}/all-last-{output_name}.csv", index=False)

    print(f"Saved results and figures to {figure_folder}")
