import numpy as np
import pandas as pd
import os
import sys
from collections import defaultdict
from utils import MetricAggregator

import matplotlib.pyplot as plt


def plot_output(
    metrics, output, labels, figure_folder_name, figure_name_suffix, dataset=None
):
    plt.rcParams.update({"font.size": 18})
    plt.rcParams["axes.labelsize"] = 20
    plt.rcParams["axes.titlesize"] = 20
    plt.rcParams["xtick.labelsize"] = 18
    plt.rcParams["ytick.labelsize"] = 18
    plt.rcParams["legend.fontsize"] = 18

    style_map = {
        "TS-MR": {"marker": 6, "color": "C0"},
        "Greedy-MR": {"marker": 4, "color": "C1"},
        "LinTS": {"marker": 5, "color": "C2"},
        "TS-Freq": {"marker": 7, "color": "C3"},
        "OFUL": {"marker": "o", "color": "C4"},
        "Greedy": {"marker": "v", "color": "C5"},
    }

    os.makedirs(figure_folder_name, exist_ok=True)
    for name, metric in metrics.items():
        if name not in ["cumuregrets", "oful_fraction", "oful_indicators"]:
            continue
            
        fig = plt.figure(figsize=(10, 6))  
        max_y = []
        for alg, agg in metric.items():
            if name == "mus" and alg not in ["LinTS", "Greedy", "TS-MR", "Greedy-MR"]:
                continue
            if name == "oful_fraction" and alg not in ["TS-MR", "Greedy-MR"]:
                continue
            if name == "oful_indicators" and alg not in ["TS-MR", "Greedy-MR"]: # old name
                continue
            style = style_map[alg]
            agg.plot(ax=plt.gca(), label=alg, marker=style["marker"], mark_num=10)
            plt.gca().lines[-1].set_color(style["color"])  # Set consistent color

            mean, se = agg.get_mean_se()
            max_y.append(np.max(mean + se))


        if name == "mus":
            plt.ylim(0, 15)
            
        plt.xlabel("Time")
        plt.ylabel(labels[name])

        plt.legend()
        figure_file_name = figure_folder_name + f"/{'oful_fraction' if name == 'oful_indicators' else name}" + figure_name_suffix
        plt.savefig(figure_file_name, dpi=600, bbox_inches="tight")
        plt.close(fig)


def plot_saved_results(output_name, T=None):
    """
    Read saved results and reproduce figures from a previous experiment run.

    Args:
        output_name: Base name of the saved output files (without .csv extension)
        t: Optional time limit for plotting. If None, plots full time range.
    """
    output_folder = "my_outputs"
    figure_folder = f"figures/figures-{output_name}"

    outputs = pd.read_csv(f"my_outputs/all-{output_name}.csv")

    # Extract parameters from output name
    params = output_name.split("_")
    param_dict = {}
    i = 0
    while i < len(params):
        if params[i] == "sim":
            param_dict["sim"] = params[i + 1]
            i += 2
        elif params[i] in ["n", "k", "t", "r"]:
            param_dict[params[i]] = (
                None if params[i + 1] == "None" else float(params[i + 1])
            )
            i += 2
        elif params[i] in ["pm", "ps", "ns", "mts", "mg"]:
            param_dict[params[i]] = (
                None if params[i + 1] == "None" else float(params[i + 1])
            )
            i += 2
        elif params[i] == "ds":
            param_dict["ds"] = "_".join(params[i + 1 : -1])  
            break
        else:
            i += 1

    metrics = defaultdict(lambda: defaultdict(MetricAggregator))
    for col in outputs.columns:
        if col.endswith("_m0"):
            metric_name = col[:-3]  # Remove '_m0' suffix
            alg = metric_name.split("_")[0]
            metric_type = "_".join(metric_name.split("_")[1:])

            agg = MetricAggregator()
            if T is not None:
                agg.m0 = outputs[metric_name + "_m0"].values[:T]
                agg.m1 = outputs[metric_name + "_m1"].values[:T]
                agg.m2 = outputs[metric_name + "_m2"].values[:T]
            else:
                agg.m0 = outputs[metric_name + "_m0"].values
                agg.m1 = outputs[metric_name + "_m1"].values
                agg.m2 = outputs[metric_name + "_m2"].values

            metrics[metric_type][alg] = agg

 
    labels = {
        "regrets": "Instantaneous Regret",
        "cumuregrets": "Cumulative Regret",
        "mus": "$\mu_t$",
        "zeta": "$\zeta_t$",
        "oful_fraction": "Fraction of OFUL Actions",
        "oful_indicators": "Fraction of OFUL Actions", # old name
    }


    figure_suffix = ".jpg"


    plot_output(
        metrics, outputs, labels, figure_folder, figure_suffix, param_dict.get("ds")
    )
