import abc
from typing import List, Callable

import numpy as np
import numpy.linalg as npl
import numpy.random as npr

from envs import Context
from envs import Feedback
from utils import DataSummary


class Metric:
    regrets: List[float]

    def __init__(self):
        self.regrets = []
        self.chosen_arms = []

    def add_obs(self, feedback: Feedback):
        self.regrets.append(feedback.regret)
        self.chosen_arms.append(feedback.regret)


class WorthFunction:
    summary: DataSummary

    def bind(self, summary: DataSummary):
        self.summary = summary
        self.update()

    def update(self):
        pass

    @abc.abstractmethod
    def compute(self, ctx: Context) -> np.ndarray:
        pass


class ProductWorthFunction(WorthFunction):
    def compute(self, ctx: Context) -> np.ndarray:
        values = ctx.arms @ self.candidates()

        if values.ndim == 2:
            values = np.max(values, axis=1)

        return values

    @abc.abstractmethod
    def candidates(self):
        pass


class Policy:
    metrics: Metric

    def __init__(self):
        self.metrics = Metric()

    @abc.abstractmethod
    def choose_arm(self, ctx: Context) -> int:
        pass

    def update(self, feedback: Feedback):
        self.update_model(feedback)
        self.update_metrics(feedback)

    @abc.abstractmethod
    def update_model(self, feedback: Feedback):
        pass

    def update_metrics(self, feedback: Feedback):
        self.metrics.add_obs(feedback)


class Roful(Policy):
    def __init__(
        self,
        d,
        lambda_,
        worth_func: WorthFunction,
        param=None,
        noise_sd=None,
        t=None,
        radius=None,
        delta=None,
        inflation=None,
    ):
        super().__init__()

        # Core parameters
        self.summary = DataSummary(d, lambda_, noise_sd, radius, param, inflation, t)
        self.worth_func = worth_func
        self.inflation = inflation if inflation else Roful.const_inflation(1.0)

        # State tracking
        self.tt = 0
        self.t = t
        self.radius = radius
        self.delta = delta
        self.param = param
        self.noise_sd = noise_sd

        # Metrics tracking
        self.zetas = []
        self.mu_indicators = []  # Track if mu is below threshold

    @staticmethod
    def ts(
        d,
        lambda_,
        inflation=1.0,
        state=None,
        param=None,
        noise_sd=None,
        t=None,
        radius=None,
        delta=None,
    ):
        if isinstance(inflation, float):
            inflation = Roful.const_inflation(inflation)
        return Roful(
            d,
            lambda_,
            TSWorthFunction(inflation, delta=delta, state=state),
            param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=delta,
            inflation=inflation,
        )

    @staticmethod
    def oful(
        d,
        lambda_,
        radius_oful,
        param=None,
        noise_sd=None,
        t=None,
        radius=None,
        delta=None,
    ):
        return Roful(
            d,
            lambda_,
            SievedGreedyWorthFunction(radius_oful, 1.0),
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=delta,
            inflation=Roful.const_inflation(1.0),
        )

    @staticmethod
    def tsmr(
        d,
        lambda_,
        inflation=1.0,
        state=None,
        param=None,
        noise_sd=None,
        t=None,
        radius=None,
        delta=None,
        mu_threshold=None,
        radius_oful=None,
    ):
        if isinstance(inflation, float):
            inflation = Roful.const_inflation(inflation)
        return Roful(
            d,
            lambda_,
            TSMRWorthFunction(
                inflation,
                mu_threshold=mu_threshold,
                radius=radius,
                noise_sd=noise_sd,
                delta=delta,
                radius_oful=radius_oful,
                tolerance=1,
                state=state,
            ),
            param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=delta,
            inflation=inflation,
        )

    @staticmethod
    def greedy(d, lambda_, param=None, noise_sd=None, t=None, radius=None, delta=None):
        return Roful(
            d,
            lambda_,
            GreedyWorthFunction(),
            param=param,
            noise_sd=noise_sd,
            t=t,
            radius=radius,
            delta=delta,
            inflation=Roful.const_inflation(1),
        )

    @staticmethod
    def const_inflation(value):
        return lambda *args: value

    @staticmethod
    def conditional_inflation(value, thin_thresh):
        return lambda summary: (
            summary.radius_det() if summary.thinness > thin_thresh else value
        )

    @staticmethod
    def radius_OFUL(delta=1e-3):
        return lambda summary: summary.radius_OFUL(delta)

    @staticmethod
    def radius_inflation(delta=1e-3):
        return lambda summary: summary.radius_OFUL(delta)

    @staticmethod
    def ts_freq_inflation(delta=1e-3):
        return lambda summary: summary.radius_normal(delta)

    @staticmethod
    def ts_inflation_alter(delta=1e-3):
        return (
            lambda summary: summary.noise_sd
            * summary.radius_normal(delta)
            / summary.radius_det(delta)
        )

    @property
    def d(self):
        return self.summary.d

    def choose_arm(self, ctx: Context) -> int:
        self.worth_func.bind(self.summary)
        values = self.worth_func.compute(ctx)
        return np.argmax(values).item()

    def exploration_amount(self, arms):
        scale = arms @ npl.solve(self.summary.xx, arms.T)
        return np.diag(scale) ** 0.5 if len(scale.shape) == 2 else scale**0.5

    def update_model(self, feedback: Feedback):
        self.summary.add_obs(feedback.chosen_arm, feedback.rew)

    def update_metrics(self, feedback: Feedback):
        super().update_metrics(feedback)
        self.tt += 1

        # Update core metrics
        lambda_max = max(self.summary.scale)
        theta_hat = self.summary.mean
        theta_norm = npl.norm(theta_hat)

        self.zetas.append(
            0.0
            if theta_norm == 0 or lambda_max == 0
            else theta_hat.T @ self.summary.xx @ theta_hat / lambda_max / theta_norm**2
        )

        self.outputs = (
            self.worth_func.mus,
            self.metrics.regrets,
            self.zetas,
            self.worth_func.oful_indicators,
        )


class TSWorthFunction(ProductWorthFunction):
    compensator: np.ndarray

    def __init__(self, inflation, delta, state=npr):
        self.inflation = inflation
        self.state = state
        self.alphas = []
        self.mus = []
        self.delta = delta
        self.discrete_alphas = []
        self.oful_indicators = []

    @property
    def d(self):
        return self.summary.d

    def update(self):
        d = self.d
        rand = self.state.randn(d)

        basis = self.summary.basis
        scale = self.summary.scale

        inv_principle_scale = np.zeros_like(scale)
        inv_principle_scale[0] = 1 / self.summary.scale[0]

        self.radius_det = self.summary.radius_det()
        self.radius_TS = self.summary.radius_TS()

        self.compensator = (
            self.radius_TS
            * basis.T
            @ (rand / scale**0.5)
            / self.summary.radius_normal(self.delta)
        )

    def candidates(self):
        return self.summary.mean + self.compensator

    def principle_candidates(self):
        return self.summary.mean + self.principle_compensator

    def compute(self, ctx: Context) -> np.ndarray:
        discrete_alpha = self.summary.calculate_discrete_alpha(ctx.arms)

        values = ctx.arms @ self.candidates()

        if values.ndim == 2:
            values = np.max(values, axis=1)

        discrete_mu = (self.summary.inflation(self.summary) + 1.0) * (
            1 + discrete_alpha
        )

        self.discrete_alphas.append(discrete_alpha)
        self.alphas.append(discrete_alpha)
        self.mus.append(discrete_mu)
        self.oful_indicators.append(False)  # Pure TS does not have OFUL action
        return values


class GreedyWorthFunction(ProductWorthFunction):
    def __init__(self, inflation=1.0):
        self.inflation = inflation
        self.alphas = []
        self.mus = []
        self.oful_indicators = []

    def candidates(self):
        return self.summary.mean

    def compute(self, ctx: Context) -> np.ndarray:
        values = ctx.arms @ self.candidates()

        if values.ndim == 2:
            values = np.max(values, axis=1)
        self.alpha_approx = self.summary.calculate_alpha
        self.mu_approx = self.summary.calculate_mu
        self.alphas.append(self.alpha_approx)
        self.mus.append(self.mu_approx)
        self.oful_indicators.append(False)  # Pure Greedy does not have OFUL action
        return values


class SievedGreedyWorthFunction(WorthFunction):
    radius: Callable
    tolerance: float

    def __init__(self, radius, tolerance):
        self.radius = radius
        self.tolerance = tolerance
        self.alphas = []
        self.mus = []
        self.discrete_alphas = []
        self.oful_indicators = []

    def compute(self, ctx: Context) -> np.ndarray:
        lowers, centers, uppers = self.confidence_bounds(ctx.arms)

        # sieving arms
        baseline = lowers.max()
        optimal = uppers.max()
        minus_infty = np.ones_like(centers) * (-np.inf)

        threshold = self.tolerance * optimal + (1.0 - self.tolerance) * baseline
        survivors = uppers >= threshold

        # computing the values

        self.radius_TS = self.summary.radius_TS()
        self.radius_det = self.summary.radius_det()
        discrete_alpha = self.summary.calculate_discrete_alpha(ctx.arms)

        discrete_mu = (self.summary.inflation(self.summary) + 1.0) * (
            1 + discrete_alpha
        )

        if self.tolerance == 1.0:
            self.mu_approx = 2.0
            self.discrete_alpha = 1.0
            self.discrete_alphas.append(self.discrete_alpha)
            self.alpha_approx = 1.0
            self.oful_indicators.append(True)  # Pure OFUL action

            self.alphas.append(self.discrete_alpha)
            self.mus.append(self.mu_approx)
        else:
            self.discrete_alpha = discrete_alpha
            self.discrete_alphas.append(self.discrete_alpha)
            self.alpha_approx = self.summary.calculate_alpha
            self.mu_approx = discrete_mu
            self.alphas.append(self.alpha_approx)
            self.mus.append(self.mu_approx)

            self.oful_indicators.append(None)  # Not applicable

        return np.where(survivors, centers, minus_infty)

    def confidence_center(self, arms):

        return arms @ self.summary.mean

    def confidence_width(self, arms):
        scale = arms @ npl.solve(self.summary.xx, arms.T)

        if len(scale.shape) == 2:
            scale = np.diag(scale)

        return self.radius(self.summary) * scale**0.5

    def confidence_bounds(self, arms):
        centers = self.confidence_center(arms)
        widths = self.confidence_width(arms)
        return centers - widths, centers, centers + widths


class TSMRWorthFunction(WorthFunction):
    compensator: np.ndarray

    def __init__(
        self,
        inflation,
        mu_threshold,
        radius,
        noise_sd,
        delta,
        radius_oful,
        tolerance,
        state=npr,
    ):
        self.inflation = inflation
        self.state = state
        self.mu_threshold = mu_threshold
        self.radius_oful = radius_oful
        self.tolerance = tolerance
        self.radius = radius
        self.noise_sd = noise_sd
        self.delta = delta
        self.alphas = []
        self.mus = []
        self.discrete_alphas = []
        self.oful_indicators = []

    @property
    def d(self):
        return self.summary.d

    @property
    def t(self):
        return self.summary.t

    def compute(self, ctx: Context) -> np.ndarray:
        values = ctx.arms @ self.candidates()

        _, _, uppers = self.confidence_bounds(ctx.arms)

        if values.ndim == 2:
            values = np.max(values, axis=1)

        discrete_alpha = self.summary.calculate_discrete_alpha(ctx.arms)
        discrete_mu = (self.summary.inflation(self.summary) + 1.0) * (
            1 + discrete_alpha
        )

        if discrete_mu < self.mu_threshold:
            self.discrete_alpha = discrete_alpha
            self.alphas.append(discrete_alpha)
            self.mu_approx = discrete_mu
            self.mus.append(self.mu_approx)
            self.oful_indicators.append(False)

            self.discrete_alphas.append(self.discrete_alpha)
            return values
        else:
            self.mu_approx = 2.0
            self.discrete_alpha = discrete_alpha
            self.oful_indicators.append(True)

            self.alphas.append(self.discrete_alpha)

            self.mus.append(self.mu_approx)

            self.discrete_alphas.append(self.discrete_alpha)

            return uppers

    def update(self):
        d = self.d
        rand = self.state.randn(d)

        basis = self.summary.basis
        scale = self.summary.scale

        inv_principle_scale = np.zeros_like(scale)
        inv_principle_scale[0] = 1 / self.summary.scale[0]

        self.radius_det = self.summary.radius_det()
        self.radius_TS = self.summary.radius_TS()
        self.compensator = (
            self.radius_TS
            * basis.T
            @ (rand / scale**0.5)
            / self.summary.radius_normal(self.delta)
        )

        self.beta = self.summary.radius_det()

    def candidates(self):

        return self.summary.mean + self.compensator

    def confidence_center(self, arms):
        return arms @ self.summary.mean

    def confidence_width(self, arms):
        scale = arms @ npl.solve(self.summary.xx, arms.T)

        if len(scale.shape) == 2:
            scale = np.diag(scale)

        return self.radius_oful(self.summary) * scale**0.5

    def confidence_bounds(self, arms):
        centers = self.confidence_center(arms)
        widths = self.confidence_width(arms)

        return centers - widths, centers, centers + widths

    def exploration_amount(self, arms):
        scale = arms @ npl.solve(self.summary.xx, arms.T)

        if len(scale.shape) == 2:
            scale = np.diag(scale)

        return scale**0.5
