import numpy as np
import numpy.linalg as npl
import numpy.random as npr
import pandas as pd
from collections import defaultdict
import os


class DataSummary:
    def __init__(self, dim, lambda_, noise_sd, param_bound, param, inflation, T):
        self.lambda_ = lambda_
        self.param_bound = param_bound
        self.noise_sd = noise_sd
        self.xy = np.zeros(dim, dtype=np.float64)
        self.xx = np.eye(dim, dtype=np.float64) * lambda_

        self._mean = np.zeros(dim, dtype=np.float64)
        self._basis = np.eye(dim, dtype=np.float64)
        self._scale = np.ones(dim, dtype=np.float64) * lambda_

        self._dirty = False
        self.t = 0
        self.param = param
        self.inflation = inflation
        self.T = T

    def _update_caches(self):
        U, s, Vt = npl.svd(self.xx, hermitian=True)
        self._mean = U @ ((Vt @ self.xy) / s)
        self._basis = Vt
        self._scale = s
        self._dirty = False

    def add_obs(self, x, y, tau=1.0):
        self.xy += x * y / tau**2
        self.xx += np.outer(x, x) / tau**2
        self.t += 1
        self._dirty = True

    @property
    def d(self):
        return self.xy.shape[0]

    @property
    def TS_prior_var(self):
        return 1.0 / self.lambda_

    @property
    def mean(self):
        if self._dirty:
            self._update_caches()
        return self._mean

    @property
    def basis(self):
        if self._dirty:
            self._update_caches()
        return self._basis

    @property
    def scale(self):
        if self._dirty:
            self._update_caches()
        return self._scale

    @property
    def thinness(self):
        if self._dirty:
            self._update_caches()
        scale_inv = 1 / self.scale
        return (max(scale_inv) * self.d / sum(scale_inv)) ** 0.5

    def radius_det(self, delta=1e-3):
        if self._dirty:
            self._update_caches()
        term1 = np.log(self.scale / self.lambda_).sum() - 2 * np.log(delta / 2 / self.T)
        term2 = self.lambda_ * self.param_bound**2
        return self.noise_sd * term1**0.5 + term2**0.5

    def radius_OFUL(self, delta=1e-3):
        return self.radius_det(delta=delta)

    def radius_TS(self, delta=1e-3):
        return self.inflation(self) * self.radius_det(delta=delta)

    def radius_normal(self, delta=1e-3):
        return np.sqrt(2 * self.d * np.log(2 * self.d / delta * 2 * self.T))

    @staticmethod
    def new_bound(scale, XX_norm_sqr_min, XX_norm_sqr_max):
        lambda_max = np.max(scale)
        lambda_min = np.min(scale)
        lambda_up = np.min(scale[scale >= XX_norm_sqr_max])
        lambda_down = np.max(scale[scale <= XX_norm_sqr_max])

        bound_min = (
            1 / lambda_max + 1 / lambda_min - XX_norm_sqr_min / lambda_max / lambda_min
        )
        bound_max = (
            1 / lambda_up + 1 / lambda_down - XX_norm_sqr_max / lambda_up / lambda_down
        )

        return bound_min, bound_max

    @property
    def calculate_alpha(self):
        _, wst_x_invnorm_sqr_TS = self.calculate_x_invnorm_sqr_range(self.radius_TS())
        opt_x_invnorm_sqr_RLS, _ = self.calculate_x_invnorm_sqr_range(self.radius_det())

        if wst_x_invnorm_sqr_TS == 0:
            return float("inf")

        return np.sqrt(opt_x_invnorm_sqr_RLS / wst_x_invnorm_sqr_TS)

    @property
    def calculate_mu(self):
        return (self.inflation(self) + 1.0) * (1 + self.calculate_alpha)

    def calculate_x_invnorm_sqr_range(self, ellipsoid_radius=None):
        if self._dirty:
            self._update_caches()

        theta_hat = self.mean
        theta_hat_Vt_norm_sqr = theta_hat.T @ self.xx @ theta_hat
        lambda_max = max(self.scale)
        lambda_min = min(self.scale)

        if ellipsoid_radius == 0.0:
            theta_hat_Vt_inv_norm_sqr = theta_hat.T @ npl.solve(self.xx, theta_hat)
            if theta_hat_Vt_inv_norm_sqr == 0.0:
                return 0.0, 0.0
            return (
                theta_hat_Vt_inv_norm_sqr / npl.norm(theta_hat) ** 2,
                theta_hat_Vt_inv_norm_sqr / npl.norm(theta_hat) ** 2,
            )

        if theta_hat_Vt_norm_sqr >= ellipsoid_radius**2:
            zeta = (theta_hat_Vt_norm_sqr - ellipsoid_radius**2) ** 0.5
            theta_norm_upper_bound = npl.norm(theta_hat) + ellipsoid_radius / np.sqrt(
                lambda_min
            )

            x_XXnorm_max = np.sqrt(theta_hat.T @ self.xx @ self.xx @ theta_hat) / zeta
            x_XXnorm_min = zeta / theta_norm_upper_bound

            return self.new_bound(
                self.scale,
                np.maximum(x_XXnorm_min**2, lambda_min),
                np.minimum(x_XXnorm_max**2, lambda_max),
            )

        return 1 / lambda_min, 1 / lambda_max

    def confidence_center(self, arms):
        return arms @ self.mean

    def confidence_width(self, arms, ellipse_radius):
        scale = arms @ npl.solve(self.xx, arms.T)
        if len(scale.shape) == 2:
            scale = np.diag(scale)
        return ellipse_radius * scale**0.5, scale

    def confidence_bounds(self, arms, ellipse_radius):
        centers = self.confidence_center(arms)
        widths, _ = self.confidence_width(arms, ellipse_radius)
        return centers - widths, centers, centers + widths

    def calculate_discrete_alpha(self, arms):
        lowers_PVT, _, uppers_PVT = self.confidence_bounds(
            arms, self.radius_TS()
        )
        lowers_RLS, _, uppers_RLS = self.confidence_bounds(
            arms, self.radius_det()
        )

        survivors_PVT = uppers_PVT >= np.max(lowers_PVT)
        survivors_RLS = uppers_RLS >= np.max(lowers_RLS)

        _, x_XXinvnorm_sqr = self.confidence_width(arms, self.radius_det())

        # Handle case where there are no survivors in PVT set
        if not np.any(survivors_PVT) or np.min(x_XXinvnorm_sqr[survivors_PVT]) == 0:
            return float("inf")

        # Handle case where there are no survivors in RLS set
        if not np.any(survivors_RLS):
            return 0.0

        return np.sqrt(np.max(x_XXinvnorm_sqr[survivors_RLS])) / np.sqrt(
            np.min(x_XXinvnorm_sqr[survivors_PVT])
        )


class MetricAggregator:
    def __init__(self):
        self.m0 = []  # count
        self.m1 = []  # sum
        self.m2 = []  # sum of squares

    def confidence_band(self):
        m0 = np.maximum(np.array(self.m0), 1)
        m1 = np.array(self.m1)
        m2 = np.array(self.m2)

        # Filter out NA values
        valid_mask = ~np.isnan(m1) & ~np.isnan(m2)
        m0 = m0[valid_mask]
        m1 = m1[valid_mask]
        m2 = m2[valid_mask]

        mean = m1 / m0
        var = np.maximum((m2 - m1**2 / m0) / (m0 - 1), 0)
        se = np.sqrt(var / m0)

        return mean, np.sqrt(var), se

    def plot(self, ax, label, marker, mark_num, scale=2.0):
        mean, _, se = self.confidence_band()
        x = np.arange(len(mean))

        ax.fill_between(x, mean - scale * se, mean + scale * se, alpha=0.2)
        ax.plot(
            x,
            mean,
            label=label,
            marker=marker,
            markevery=int(len(x) / mark_num),
            markersize=5,
        )

    def get_mean_se(self):
        mean, _, se = self.confidence_band()
        return mean, se

    def aggregate(self, xs, filter=lambda _: True):
        self._ensure_len(len(xs))
        for i, x in enumerate(xs):
            if filter(i):
                self.m0[i] += 1
                self.m1[i] += x
                self.m2[i] += x**2

    def _ensure_len(self, n):
        if len(self.m0) < n:
            extension = [0] * (n - len(self.m0))
            self.m0.extend(extension)
            self.m1.extend(extension)
            self.m2.extend(extension)


class save_results:
    def __init__(self):
        self.outputs = []
        self.outputs_last = []
        self.labels = {
            "regrets": "Instantaneous Regret",
            "cumuregrets": "Cumulative Regret",
            "mus": "$\mu_t$",
            "zeta": "$\zeta_t$",
            "oful_fraction": "Fraction of OFUL Actions",
        }

    def init_outputs(self):

        self.metrics = {
            name: defaultdict(MetricAggregator)
            for name in ["regrets", "cumuregrets", "mus", "zeta", "oful_fraction"]
        }

    def aggregate_metrics(self, results):
        for name, metrics in results.items():
            (mu, regret, zeta, oful_fraction) = metrics

            self.metrics["regrets"][name].aggregate(regret)
            self.metrics["cumuregrets"][name].aggregate(np.cumsum(regret))

            self.metrics["mus"][name].aggregate(mu)

            self.metrics["zeta"][name].aggregate(zeta)
            self.metrics["oful_fraction"][name].aggregate(oful_fraction)

    def aggregate_output(self):
        # Build dictionaries to collect all data first
        output_data = {}
        output_last_data = {}

        for metric_name, metric in self.metrics.items():
            for alg, agg in metric.items():
                mean, se = agg.get_mean_se()
                nm = f"{alg}_{metric_name}"

                output_data[f"{nm}_mean"] = mean
                output_data[f"{nm}_se"] = se
                output_data[f"{nm}_m0"] = agg.m0
                output_data[f"{nm}_m1"] = agg.m1
                output_data[f"{nm}_m2"] = agg.m2

                output_last_data[f"{nm}_mean"] = [mean[-1]]
                output_last_data[f"{nm}_se"] = [se[-1]]
                output_last_data[f"{nm}_m0"] = [agg.m0[-1]]
                output_last_data[f"{nm}_m1"] = [agg.m1[-1]]
                output_last_data[f"{nm}_m2"] = [agg.m2[-1]]

        # Create DataFrames all at once
        output = pd.DataFrame(output_data)
        output_last = pd.DataFrame(output_last_data)

        self.outputs.append(output)
        self.outputs_last.append(output_last)

        return self.metrics, output, output_last

    def save_outputs(self, output_folder_name, output_name):
        os.makedirs(output_folder_name, exist_ok=True)

        outputs = pd.concat(self.outputs)
        outputs_last = pd.concat(self.outputs_last)

        outputs.to_csv(f"{output_folder_name}/all-{output_name}.csv", index=False)
        outputs_last.to_csv(
            f"{output_folder_name}/all-last-{output_name}.csv", index=False
        )

        return outputs, outputs_last


class StateFactory:
    def __init__(self, seed):
        self.seed = seed

    def __call__(self):
        state = npr.RandomState(self.seed)
        self.seed += 1
        return state
